import { WebElement } from '..'

interface ComponentConfig {
	tag: `${string}-${string}`
	attrs?: Record<string, AttributeOptions>
	shadow?: boolean
}

/**
 * Decorator to define a custom element.
 * @param tagOrConfig The tag name of the custom element. it MUST respect the formatting else the browser will not recognize it.
 */
export function Component(tagOrConfig: ComponentConfig['tag'] | Omit<ComponentConfig, 'attrs'>) {
	const conf = typeof tagOrConfig === 'string' ? { tag: tagOrConfig } : tagOrConfig
	return function(constructor: typeof WebElement) {

		const child = class extends constructor {
			// public static override readonly tag = tag
			public constructor() {
				super()
				// eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
				const proto = ((this as any).__proto__ as typeof WebElement)
				Object.assign(proto.__config, conf)
				proto.__config.tag = conf.tag
				const config = proto.__config
				for (const attr of Object.keys(config.attrs ?? {})) {
					const pouet = config.attrs![attr]
					// console.log(attr, pouet)
					// eslint-disable-next-line @typescript-eslint/no-dynamic-delete
					delete proto[attr as keyof typeof WebElement]
					lateAttribute(this, attr, pouet)
				}
			}
		}

		if (typeof customElements === 'undefined') {
			// eslint-disable-next-line @typescript-eslint/no-unsafe-return
			return child as any
		}

		if (!customElements.get(conf.tag)) {
			customElements.define(conf.tag, child)
		} else {
			throw new Error(`Custom element ${conf.tag} is already defined`)
		}
	}
}

export interface AttributeOptions {
	type?: 'string' | 'number' | 'boolean'
	onChange?: (newValue: any, oldValue: any) => void
}

export function lateAttribute(el: WebElement, name: string, options: AttributeOptions = {}) {
	if (!(el instanceof WebElement)) {
		return
	}

	const privateKey = `__${name}`


	Object.defineProperty(el, name, {
		get(this: WebElement) {
			return this[privateKey as keyof WebElement]
		},
		set(this: WebElement, newVal) {
			const oldVal = this[privateKey as keyof WebElement]

			const changed = newVal !== oldVal
			if (!changed) {
				return
			}

			// eslint-disable-next-line @typescript-eslint/no-unsafe-assignment, @typescript-eslint/no-unsafe-member-access
			(this as any)[privateKey] = newVal

			options.onChange?.(newVal, oldVal)

			if (newVal == null || newVal === false) {
				this.removeAttribute(name)
			} else {
				this.setAttribute(name, String(newVal))
			}

			// invalide current rendering
			this.invalidate()

			// trigger new render
			void this.triggerRender()
		},
		configurable: true,
		enumerable: true,
	})
}

type AttributeFunction = (proto: any, name: string) => void

/**
 * Decorator to define an attribute on a WebElement.
 */
export function attribute(): AttributeFunction
/**
 * Decorator to define an attribute on a WebElement.
 * @param type The type of the attribute.
 */
// eslint-disable-next-line @typescript-eslint/unified-signatures
export function attribute(type: AttributeOptions['type']): AttributeFunction
/**
 * Decorator to define an attribute on a WebElement.
 * @param options The options for the attribute.
 */
// eslint-disable-next-line @typescript-eslint/unified-signatures
export function attribute(options: AttributeOptions): AttributeFunction
export function attribute(typeOrOptions?: AttributeOptions['type'] | AttributeOptions): AttributeFunction {
	return function(proto: typeof WebElement, name: string) {
		// eslint-disable-next-line @typescript-eslint/no-unnecessary-condition
		proto.__config ??= {} as unknown as typeof proto['__config']
		proto.__config.attrs ??= {}
		proto.__config.attrs[name] = typeof typeOrOptions === 'string' ? { type: typeOrOptions } : typeOrOptions ?? {}
	}
}
