/* eslint-disable max-depth */
/* eslint-disable complexity */
import { objectMap } from '@dzeio/object-util'
import { WebElement } from '..'

export function HTMLtoString(tag: SimpleHTMLElement): string {
	// compile attributes (note: the attributes values are encoded to not break json)
	const attrs = objectMap(tag.attrs ?? {}, (value, key) => ` ${key}="${value.replaceAll('"', '&#34;')}"`).join('')
	// console.log('a', tag)
	// console.log('a', tag.childs)
	let childs = ''
	if (tag.childs) {
		console.log('na', tag, tag.childs)
		childs = tag.childs.map((it) => typeof it === 'string' ? it : HTMLtoString(it)).join('')
	}

	return `<${tag.name}${attrs}>${childs}</${tag.name}>`
}

export function HTMLToElement(tag: SimpleHTMLElement): HTMLElement {
	// create root element
	const el = document.createElement(tag.name)

	// apply attributes
	if (tag.attrs) {
		for (const [k, v] of Object.entries(tag.attrs)) {
			el.setAttribute(k, v)
		}
	}

	// apply event handlers
	if (tag.events) {
		for (const [evt, handler] of Object.entries(tag.events)) {
			el.addEventListener(evt, handler)
		}
	}

	// apply childs
	if (tag.childs) {
		for (const child of tag.childs) {
			if (typeof child === 'string') {
				el.appendChild(document.createTextNode(child))
			} else {
				el.appendChild(HTMLToElement(child))
			}
		}
	}

	return el
}


export interface SimpleHTMLElement {
	name: string
	attrs?: Record<string, string>
	events?: Record<string, EventListener>
	childs?: Array<SimpleHTMLElement | string>
}

export function hasEvents(el: SimpleHTMLElement): boolean {
	if (Object.keys(el.events ?? {}).length > 0) {
		return true
	}
	if (el.childs) {
		for (const child of el.childs) {
			if (typeof child === 'string') {
				continue
			}
			if (hasEvents(child)) {
				return true
			}
		}
	}
	return false
}

function parseAttrs(attrString: string, values: Array<unknown>): { attrs: Record<string, string>, events: Record<string, EventListener> } {
	const attrs: Record<string, string> = {}
	const events: Record<string, EventListener> = {}
	const parts = attrString.match(/(?:[^.*=]+=(?:"[^"]*"|'[^']*'|[^.*"']+))/g) ?? []
	for (const part of parts) {
		if (!part) {
			continue
		}
		const [key, raw] = part.trim().split('=') as [string, string]
		if (!raw) {
			continue
		}
		const val = raw.replace(/^['"]?|['"]?$/g, '')
		const match = /^\u0000(\d+)\u0000$/.exec(val)
		if (match) {
			const idx = Number(match[1])
			const value = values[idx]
			if (key.startsWith('on') && typeof value === 'function') {
				events[key.slice(2)] = value as EventListener
			} else if (typeof value === 'string') {
				attrs[key] = value
			}
		} else {
			attrs[key] = val
		}
	}
	return { attrs, events }
}

// TODO: rework
function parseElement(fragment: string, values: Array<unknown> = []): SimpleHTMLElement | string {
	fragment = fragment.trim()
	if (!fragment.startsWith('<')) { return fragment }

	// self-closing tag
	const selfClose = /^<([a-zA-Z0-9_-]+)([^>]*)\/>$/.exec(fragment)
	if (selfClose) {
		const [, tagName, rawAttrs] = selfClose
		const { attrs, events } = parseAttrs(rawAttrs, values)
		return {
			name: tagName,
			attrs: Object.keys(attrs).length ? attrs : undefined,
			events: Object.keys(events).length ? events : undefined,
		}
	}

	// regular tag
	const fullMatch = /^<([a-zA-Z0-9_-]+)([^>]*)>([\s\S]*)<\/\1>$/.exec(fragment)
	if (!fullMatch) { return fragment }

	const [, tagName, rawAttrs, inner] = fullMatch
	const { attrs, events } = parseAttrs(rawAttrs, values)
	const childs: Array<SimpleHTMLElement | string> = []

	let rest = inner.trim()

	while (rest) {
		// interpolation placeholder
		const interp = /^\u0000(\d+)\u0000/.exec(rest)
		if (interp) {
			const idx = Number(interp[1])
			const val = values[idx]
			if (val instanceof WebElement) {
				// childs.push(val.render())
			} else if (typeof val === 'object' && val !== null && 'name' in val) { childs.push(val as SimpleHTMLElement) }
			else { childs.push(String(val)) }
			rest = rest.slice(interp[0].length).trim()
			continue
		}

		if (rest.startsWith('<')) {
			const tagMatch = /^<([a-zA-Z0-9_-]+)/.exec(rest)
			if (!tagMatch) { break }
			const childTag = tagMatch[1]
			let depth = 0
			let i = 0
			for (; i < rest.length; i++) {
				if (rest.startsWith(`<${childTag}`, i)) { depth++ }
				else if (rest.startsWith(`</${childTag}>`, i)) {
					depth--
					if (depth === 0) {
						i += (`</${childTag}>`).length
						break
					}
				}
			}
			const chunk = rest.slice(0, i)
			childs.push(parseElement(chunk, values) as SimpleHTMLElement)
			rest = rest.slice(i).trim()
		} else {
			const nextIndices = [rest.indexOf('<'), rest.indexOf('\u0000')].filter((i) => i >= 0)
			const idx = nextIndices.length ? Math.min(...nextIndices) : -1
			const text = idx >= 0 ? rest.slice(0, idx) : rest
			childs.push(text.trim())
			rest = idx >= 0 ? rest.slice(idx).trim() : ''
		}
	}

	return {
		name: tagName,
		attrs: Object.keys(attrs).length ? attrs : undefined,
		events: Object.keys(events).length ? events : undefined,
		childs: childs.length ? childs : undefined,
	}
}

/**
 * Parses a tagged template literal into a SimpleHTMLElement.
 */
export function html(str: string): SimpleHTMLElement
export function html(strings: TemplateStringsArray, ...values: Array<string | number | undefined | boolean | object | null | WebElement>): SimpleHTMLElement
export function html(strings: TemplateStringsArray | string, ...values: Array<unknown>): SimpleHTMLElement {
	// input as a raw string, limited parsing.
	if (typeof strings === 'string') {
		return parseElement(strings.replaceAll('&#34;', '"')) as SimpleHTMLElement
	}

	// the new built string
	let full = ''

	// indexes of values that are already parsed into the `full` string
	for (let idx = 0; idx < strings.length; idx++) {
		full += strings[idx]!
		if (idx < values.length) {
			const value = values[idx]
			switch (typeof value) {
				case 'undefined': {
					break
				}
				// if the value is a string, add it to the full string and mark it for removal
				case 'string': {
					full += values[idx] as string
					break
				}
				case 'object': {
					if (value === null) {
						break
					}
					if (value instanceof WebElement) {

					} else {
						const attrs = Object.entries(value as Record<string, unknown>)
							.filter(([, value]) => typeof value !== 'undefined' && value !== null)
							.map(([key, value]) => `${key}="${(value as string).replace('"', '&#34;')}"`)
							.join(' ')
						full += attrs
						break
					}
				}
				default: {
					// add a placeholder for the value to be parsed later
					full += `\u0000${idx}\u0000`
				}
			}
		}
	}

	// parse & return :D
	return parseElement(full.replaceAll('&#34;', '"'), values) as SimpleHTMLElement
}

type ClassList = Array<string | Record<string, any>>

/**
 * Simple helper function to create a string with class names.
 * @param items - Array of class names or objects with class names as keys and boolean values as values.
 * @returns A string with the class names separated by spaces.
 */
export function cls(items: string | ClassList): string
/**
 * Simple helper function to create a string with class names.
 * @param items - Array of class names or objects with class names as keys and boolean values as values.
 * @returns A string with the class names separated by spaces.
 */
export function cls(...items: ClassList | Array<ClassList>): string
/**
 * Simple helper function to create a string with class names.
 * @param items - Array of class names or objects with class names as keys and boolean values as values.
 * @returns A string with the class names separated by spaces.
 */
export function cls(...items: Array<string | undefined | null | Record<string, any>>): string {
	if (items.length === 1 && typeof items[0] === 'string') {
		return items[0]
	} else if (items.length === 1 && Array.isArray(items[0])) {
		items = items[0] as ClassList
	}

	return items.map((item) => {
		if (typeof item === 'undefined' || item === null) {
			return null
		}
		if (typeof item === 'string') {
			return item
		}
		return Object.keys(item).filter((key) => item[key]).join(' ')
	}).filter((it) => !!it).join(' ')
}

export function attrs(string: Record<string, any>): string {
	return ''
}

export function getContext(): 'browser' | 'node' {
	if (typeof document !== 'undefined') {
		return 'browser'
	}
	return 'node'
}


export function assert(bool: any, message?: string): asserts bool {
	if (!bool) {
		throw new Error(message ?? 'Assertion failed')
	}
}
