/* eslint-disable max-classes-per-file */

import { objectLoop } from '@dzeio/object-util'

import { assert, type AttributeOptions, html, type SimpleHTMLElement, HTMLToElement, HTMLtoString, getContext, hasEvents } from '..'

// Polyfill HTMLElement on server to allow SSR rendering
let localHTMLElement: typeof HTMLElement
if (typeof HTMLElement !== 'undefined') {
	localHTMLElement = HTMLElement
} else {
	// @ts-expect-error polyfill for SSR
	localHTMLElement = class HTMLElement {
		public attachShadow(_options: ShadowRootInit): ShadowRoot {
			return undefined as unknown as ShadowRoot
		}
		public setAttribute(_name: string, _value: string) {
			// super.setAttribute(name, value)
		}
	}
}

/**
 * Wrapper around HTMLElement that provides a simple way to create custom elements.
 */
export default class WebElement extends localHTMLElement {
	public static readonly tag: string = 'web-element'

	public static __config: {
		tag: string
		attrs?: Record<string, AttributeOptions>
	} = {
		tag: 'web-element'
	}

	/**
	 * childs of the current element that will be rendered inside `<slot />` elements
	 */
	public childs: Record<string, SimpleHTMLElement | string> = {}

	/**
	 * Indicates whether the element has been mounted in the DOM.
	 */
	private mounted = false

	/**
	 * Indicates whether the element needs to be rendered.
	 */
	private needRender = true

	/**
	 * The last rendered element for caching purposes.
	 */
	private lastRender: SimpleHTMLElement | undefined

	public constructor() {
		super()

		// attach to shadow root
		// this.attachShadow({ mode: 'open' })
		// this.connectedCallback()
	}

	public static async ref(props?: any): Promise<SimpleHTMLElement> {
		const conf = this.__config
		const self = new this()
			.setProps(props)
		return html`<${conf.tag}>${(await self.renderString()).output}</${conf.tag}>`
	}

	public getConfig(): typeof WebElement['__config'] {
		return this.__proto__.__config
	}
	/**
	 * Function run when the element is connected to the DOM.
	 */
	public connectedCallback() {
		// setup slots
		if (this.dataset.slots) {
			this.setSlots(JSON.parse(this.dataset.slots) as Record<string, SimpleHTMLElement>)
			this.removeAttribute('data-slots')
		}

		// move the element inside of the shadow root
		// this.shadowRoot!.innerHTML = this.innerHTML
		// this.innerHTML = ''

		// eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
		const attrSettings = ((this as any).__proto__ as typeof WebElement).__config.attrs

		for (const key of Object.keys(attrSettings)) {
			if (this.hasAttribute(key)) {
				this.setProps({[key]: this.getAttribute(key)})
			}
		}

		// setup attributes observer
		const observer = new MutationObserver((mutations) => {

			mutations.forEach((mutation) => {
				const attr = mutation.attributeName
				if (mutation.type === 'attributes' && attr) {
					// eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
					const attrSettings = ((this as any).__proto__ as typeof WebElement).__config.attrs?.[attr]
					if (!attrSettings) {
						return
					}

					const type = attrSettings.type ?? 'string'

					switch (type) {
						case 'boolean': {
							// eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
							(this as any)[attr] = this.hasAttribute(attr) ? this.getAttribute(attr) !== 'false' : false
							break
						}
						case 'number': {
							// eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
							(this as any)[attr] = this.hasAttribute(attr) ? parseInt(this.getAttribute(attr) ?? '0', 10) : undefined
							break
						}
						default: {
							// eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
							(this as any)[attr] = this.getAttribute(attr)
						}
					}

				}
			})
		})

		observer.observe(this, {
			attributes: true
		})

		// trigger frontend hydration
		if (this.dataset.hydrate === 'true') {
			this.removeAttribute('data-hydrate')
			void this.triggerRender()
		}
	}

	/**
	 * Called when the component is mounted in the DOM
	 */
	public didMount() { /** child to implement */}

	/**
	 * Called when the component was updated from an attribute change
	 */
	public didUpdate() { /** child to implement */}

	/**
	 * Render the component, this MUST be stateless and return the result of the HTML `html` helper
	 *
	 * note: if a parent changes, this render function won't be run again
	 */
	public async render(): Promise<SimpleHTMLElement> {
		return html`<div />`
	}

	/**
	 * Invalidate the component, this will trigger a re-render on the next call to `triggerRender`
	 */
	public invalidate() {
		this.needRender = true
	}

	public setSlots(name: string, value: SimpleHTMLElement | string | null): this
	public setSlots(slots: Record<string, SimpleHTMLElement | string | null>): this
	public setSlots(slots: Record<string, SimpleHTMLElement | string | null> | string, element?: SimpleHTMLElement | string | null) {
		if (typeof slots === 'string') {
			slots = { [slots]: element as SimpleHTMLElement | string | null }
		}
		objectLoop(slots, (value, key) => {
			if (value === null) {
				// eslint-disable-next-line @typescript-eslint/no-dynamic-delete
				delete this.childs[key]
			} else {
				this.childs[key] = value
			}
		})
		return this
	}

	public getProps(): Record<string, any> {
		const props: Record<string, any> = {}
		for (const key of Object.keys(this.getConfig().attrs ?? {})) {
			props[key] = this[key]
		}
		return props
	}

	public setProps(props?: Partial<Record<string, any>>) {
		if (!props) {return this}
		objectLoop(props, (value, key) => {
			this[key] = value
		})

		return this
	}

	/**
	 * Render the component to a string.
	 *
	 * warning: events are excluded from the output.
	 */
	public async renderString(): Promise<{output: string, needHydration: boolean}> {
		const res = await this.renderOrCache()

		// SSR: manually replace <slot> tags
		let out = HTMLtoString(res)

		for (const [name, el] of Object.entries(this.childs)) {
			let regex: RegExp
			if (name === 'default') {
				regex = /<slot(\s+name=["']?default["']?)?\s*\/?>.*?<\/slot>|<slot(\s+name=["']?default["']?)?\s*\/?>/i
			} else {
				regex = new RegExp(
					`<slot\\s+name=["']?${name}["']?\\s*\\/?>.*?<\\/slot>|<slot\\s+name=["']?${name}["']?\\s*\\/?>`,
					'i'
				)
			}

			out = out.replace(regex, typeof el === 'string' ? el : HTMLtoString(el))
		}

		return { output: out, needHydration: hasEvents(res) }
	}

	/**
	 * Render the component to an HTMLElement for the browser.
	 *
	 * @throws {Error} if run outside the browser
	 */
	public async renderHTML(): Promise<HTMLElement> {
		assert(getContext() === 'browser', 'renderHTML can only be run inside the browser')

		const rendered = await this.renderOrCache()

		const root = HTMLToElement(rendered)
		for (const [name, el] of Object.entries(this.childs)) {
			let slot: HTMLSlotElement | undefined | null
			if (name === 'default') {
				slot = root.querySelector('slot')
			} else {
				slot = root.querySelector<HTMLSlotElement>(`slot[name="${name}"]`)
			}
			if (slot) {
				slot.replaceWith(typeof el === 'string' ? el : HTMLToElement(el))
			}
		}

		return root
	}

	/**
	 * Tell the component to render itself.
	 *
	 * *works only in browser*
	 */
	public async triggerRender() {
		// ignore automatics renders inside if we are not in the server
		if (getContext() !== 'browser' /* || !this.shadowRoot*/) {
			return
		}

		const rendered = await this.renderHTML()

		this.innerHTML = ''
		this.appendChild(rendered)

		if (!this.mounted) {
			this.mounted = true
			this.didMount()
		} else {
			this.didUpdate()
		}
	}

	private async renderOrCache(): Promise<SimpleHTMLElement> {
		if (this.needRender || !this.lastRender) {
			this.lastRender = await this.render()
		}
		return this.lastRender
	}
}
